unit MainForm;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Buttons, ComCtrls, ExtCtrls, Spin,
  // Windows-specific stuff
  Windows, MMSystem, mmErrMsg;

const
  sample_rate = 8000;
  sine_table_samples = 1 shl 15;     // number of samples in sine table
  max_buffer_samples = 16000;        // reasonable size of output buffer (< 64K)
  open_error = 'Error opening waveform audio!';
  mem_error = 'Error allocating memory!';

type
  audio_sample = -32767..32767;       // for 16-bit audio

type
  PSineTable = ^TSineTable;          // sine value store
  TSineTable = array [0..sine_table_samples-1] of audio_sample;

	PBuffer = ^TBuffer;                // output buffer type
  TBuffer = array [0..max_buffer_samples-1] of audio_sample;

type

  { TMorseMain }

  TMorseMain = class(TForm)
    cbSpacing: TComboBox;
    cbMode: TComboBox;
    edUrCall: TEdit;
    edReply: TEdit;
    GroupBox1: TGroupBox;
    StatusPanel1: TPanel;
    StatusPanel2: TPanel;
    txReply: TLabel;
    Label10: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    seSpeed: TSpinEdit;
    sePitch: TSpinEdit;
    tbVolume: TTrackBar;
    Timer: TTimer;
    Tune: TButton;
    lbLog: TListBox;
    procedure edReplyKeyPress(Sender: TObject; var Key: char);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Panel2Click(Sender: TObject);
    procedure SpeedChange(Sender: TObject);
		procedure SpacingChange(Sender: TObject);
    procedure PitchChange(Sender: TObject);
    procedure TimerTimer(Sender: TObject);
    procedure TuneClick(Sender: TObject);
		procedure VolumeChange(Sender: TObject);
    procedure WaveOutProc(hwo: HWAVEOUT; uMsg: UINT; dwParam1, dwParam2: DWORD);
  private
    { private declarations }
    angle: integer;          // current sine wave angle
    sine_table: PSineTable;  // sine-wave values are pre-stored in this array
    hWave_hdr1: PWaveHdr;   // wave headers
		hBuffer1: PWaveHdr;
    buffer_bytes: integer;   // max number of bytes in each output buffer
    chars_sent, chars_played: integer;
		f: extended;
		hWave_out: HWaveOut;     // handle to wave out device
    pcm: TWaveFormatEx;      // wave format descriptor
		key_down: boolean;
    procedure StopSending;
    procedure StartSending;
    procedure do_sine_table;
    procedure fill_buffer_with_sinewave (bfr: PBuffer;  var index, samples: integer);
		procedure fill_buffer_with_silence (bfr: pBuffer;  var index, samples: integer);
		function fill_buffer_with_char (bfr: pBuffer; ch: char): integer;
    procedure write_next_buffer (header: PWaveHdr);
  public
    { public declarations }
  end; 

var
  MorseMain: TMorseMain;

implementation

uses Arrays, Config;

{ TMorseMain }

type
	TMode = (calling, inqso, ok, listeningforcall, listeningforexch, closing, tuning);

var
	call, callstr, exchange, exch2, exchangestr, codestr: string;
	worked: string;
	dotlength, spacing, codeptr, counter: integer;
	mode: TMode;
  pmode: Integer;
	timestarted: TDateTime;
	vol: Extended;

const
  sConfigRoot = 'Settings';

function Rand(l:longint): longint;
begin
  try
    Result := Random(l);
  except
    Result := 0;
  end;
end;

function CompareNumText( s1, s2: string): boolean;
begin
  Result := StrToIntDef(s1,0) = StrToIntDef(s2,-1);
end;
 
procedure waveOutPrc(hwo: HWAVEOUT; uMsg: UINT; dwInstance, dwParam1, dwParam2: DWORD); stdcall;
begin
  MorseMain.WaveOutProc(hwo, uMsg, dwParam1, dwParam2)
end;

procedure TMorseMain.FormCreate(Sender: TObject);
var
  open_status: MMRESULT;
  i: DWord;
begin
  // disable the maximize button
  i := GetWindowLong(Self.Handle, GWL_STYLE);
  i := i and not (WS_MAXIMIZEBOX);
  i := SetWindowLong(Self.Handle, GWL_STYLE, i);
  // set the configuration file
  SetConfigFileName(ChangeFileExt(Application.ExeName,'.ini'));
  // set form position and persistent values
	GetFormPositionFromConfigFile(sConfigRoot,'WindowPos',MorseMain);
	seSpeed.Value := GetValueFromConfigFile(sConfigRoot,'Speed',20);
	cbSpacing.ItemIndex := GetValueFromConfigFile(sConfigRoot,'Spacing',0);
  pmode := GetValueFromConfigFile(sConfigRoot,'Mode',0);
	cbMode.ItemIndex := pmode;
	sePitch.Value := GetValueFromConfigFile(sConfigRoot,'Pitch',700);
	tbVolume.Position := GetValueFromConfigFile(sConfigRoot,'Volume',1);
	edUrCall.Text := GetValueFromConfigFile(sConfigRoot,'Call','MYCALL');
 
	// get the memory required for wave header
  GetMem(hWave_hdr1, SizeOf(WaveHdr));

	// estimate of reasonable output buffer size
	buffer_bytes := 2 * max_buffer_samples;

	// get the memory required for output buffer
  GetMem(hBuffer1, buffer_bytes);

	hWave_out := 0;
	// get the memory for the sine-wave table and build it
	GetMem(sine_table, SizeOf(TSineTable));
	do_sine_table;

	// set other state variables
	key_down := false;
	f := 1000;

	Randomize;
 
	SpeedChange(Sender);
	SpacingChange(Sender);
	PitchChange(Sender);
	VolumeChange(Sender);
	codestr := '';
	codeptr := 0;
	timestarted := Now;
 
  // fill in the TWaveFormatEx structure with our wave details
  with pcm do
  begin
    wFormatTag := wave_Format_PCM;         // it's PCM data
    nChannels := 1;                        // mono
    nSamplesPerSec := sample_rate;         // set the 44.1KHz rate
    nAvgBytesPerSec := 2 * sample_rate;    // two bytes per sample
    nBlockAlign := 2;                      // for mono 16-bit audio
    wBitsPerSample := 16;                  // 16-bit audio
    cbSize := 0;
  end;

  // try and open the wave device for our format of wave data
  open_status := waveOutOpen(@hWave_out, WAVE_MAPPER, @pcm, DWORD(@WaveOutPrc), Integer(Self), CALLBACK_FUNCTION);
  if open_status <> MMSYSERR_NOERROR then
  begin
    Tune.Enabled := false;
    hWave_out := 0;
    MessageDlg (open_error + #13#10 + translate_mm_error(open_status), mtWarning, [mbOK], 0);
  end;

end;

procedure TMorseMain.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  StopSending;
end;

procedure TMorseMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
	StopSending;
	if (lbLog.Items.Count > 0) and (pmode < 2) then
	begin
		if MessageDlg('Contest over! Save log to file?',mtConfirmation,[mbYes,mbNo],0) = mrYes then
		begin
			lbLog.Items.Insert(0,Caption);
			lbLog.Items.Insert(1,'');
			lbLog.Items.Insert(2,'Log for '+edUrCall.Text+' on '+FormatDateTime('d mmm yyyy hh:nn',timestarted));
			lbLog.Items.Insert(3,'');
			lbLog.Items.Add('');
			lbLog.Items.Add(StatusPanel1.Caption);
			lbLog.Items.SaveToFile(ExtractFilePath(ParamStr(0))+FormatDateTime('yyyymmddhhnn',timestarted)+'.LOG');
		end;
	end;
end;

procedure TMorseMain.edReplyKeyPress(Sender: TObject; var Key: char);
begin
  if Key = ' ' then Key := '?'
  else if Key = #13 then Key := #0;
end;

procedure TMorseMain.FormDestroy(Sender: TObject);
begin
  WriteFormPositionToConfigFile(sConfigRoot,'WindowPos',MorseMain);
	SaveValueToConfigFile(sConfigRoot,'Speed',seSpeed.Value);
	SaveValueToConfigFile(sConfigRoot,'Spacing',cbSpacing.ItemIndex);
	SaveValueToConfigFile(sConfigRoot,'Mode',pmode);
	SaveValueToConfigFile(sConfigRoot,'Pitch',sePitch.Value);
	SaveValueToConfigFile(sConfigRoot,'Volume',tbVolume.Position);
	SaveValueToConfigFile(sConfigRoot,'Call',PChar(edUrCall.Text));
	FreeMem (sine_table, SizeOf(TSineTable));
	FreeMem(hBuffer1, buffer_bytes);
	FreeMem(hWave_hdr1, SizeOf(WaveHdr));
  if hWave_out <> 0 then
    waveOutClose(hWave_out);
end;

procedure TMorseMain.FormShow(Sender: TObject);
begin
  edReply.SetFocus;
end;

procedure TMorseMain.Panel2Click(Sender: TObject);
begin
  ShellExecute(0,'open','http://www.g4ilo.com', '', '', SW_SHOWNORMAL);
end;

procedure TMorseMain.StopSending;
begin
  codestr := '';
	// is a sweep running?  if so, stop it
	if key_down then
	begin
    repeat
			Application.ProcessMessages;
    until not key_down;
  end
end;

procedure TMorseMain.StartSending;
begin
  if key_down then Exit;

  angle := 0;

	// Start by writing two buffers of data to the wave device.
  // populate the first wave header
  with hWave_hdr1^ do
  begin
    lpData := pChar(hBuffer1);     // pointer to the data
    dwBufferLength := 0;           // fill in size later
    dwBytesRecorded := 0;
    dwUser := 0;
    dwFlags := 0;
    dwLoops := 1;                  // just a single loop
    lpNext := nil;
    reserved := 0;
  end;
  hWave_hdr1^.dwBufferLength := buffer_bytes;     // actual buffer sizes
  // now fill the buffer and write it to the wave output
  write_next_buffer (hWave_hdr1);
  // ready to go
	key_down := true;
 
end;

procedure TMorseMain.SpeedChange(Sender: TObject);
var
	s: Integer;
begin
	s := seSpeed.Value;
	dotlength := ((sample_rate * 12) div 10) div s;
end;

procedure TMorseMain.SpacingChange(Sender: TObject);
begin
	spacing := cbSpacing.ItemIndex * 3;
end;

procedure TMorseMain.PitchChange(Sender: TObject);
begin
	f := sePitch.Value;
end;

procedure TMorseMain.TimerTimer(Sender: TObject);
var
	n, minsleft: Integer;
	s: string;
	timenow: TDateTime;

  function RAlign(const s: string; w: integer): string;
  var
  	l: integer;
  begin
  	l := Length(s);
  	Result := Copy('                '+s,l+17-w,w);
  end;

  function LAlign(const s: string; w: integer): string;
  begin
  	Result := Copy(s+'                ',1,w);
  end;

begin
  if mode = tuning then
    TuneClick(sender);
	if mode = listeningforcall then
	begin
		if CompareText(edReply.Text,call) = 0 then
		begin
			edReply.Text := '';
			if Pos(call+'.',worked) > 0 then
			begin
				codestr := '  '+edUrCall.Text+' SRI OM QSO BEFORE GL   '+callstr;
				mode := calling;
        txReply.Caption := 'Call >';
			end
			else
			begin
				codestr := '  '+exchangestr;
				mode := inqso;
        txReply.Caption := 'Serial >';
        if pmode = 2 then
          edReply.Text := exchange;
			end;
			codeptr := 1;
			StartSending;
		end
		else if Pos('?',edReply.Text) > 0 then
		begin
			s := edReply.Text;
			Delete(s,Pos('?',edReply.Text),1);
			edReply.Text := s;
      edReply.SelStart := Length(s);
      edReply.SelLength := 0;
			StopSending;
			case Rand(2) of
			0:	codestr := '  '+call+' ';
			1:	codestr := '  '+call+' '+call+' ';
			end;
			mode := calling;
			codeptr := 1;
      txReply.Caption := 'Call >';
			StartSending;
		end
		else if counter = 0 then
		begin
			if edreply.Text = '' then
			begin
				codestr := callstr;
				codeptr := 1;
				mode := calling;
				StartSending;
			end
			else
			begin
				if Rand(3) < 2 then
					codestr := callstr
				else
					codestr := '? ';
				codeptr := 1;
				mode := calling;
				StartSending;
			end;
      txReply.Caption := 'Call >';
		end;
	end
	else if mode = listeningforexch then
	begin
		if CompareNumText(exchange,edReply.Text) then
		begin
			edReply.Text := '';
			timenow := Now;
			minsleft := 60 - Trunc((timenow - timestarted) * 1440);
			n := lbLog.Items.Count+1;
			lbLog.Items.Add(RAlign(IntToStr(n),4)+' '+FormatDateTime('t',timenow)+' '+LAlign(call,10)+' 599 599 '+exchange);
      if n > 10 then
        lbLog.TopIndex := (n-10);
			StatusPanel1.Caption := IntToStr(n)+' contacts. Rate: '+IntToStr(Round(n / (timenow - timestarted) / 24.0))+' contacts / hr. Time remaining: '+IntToStr(minsleft)+' mins.';
			worked := worked + call + '.';
			if minsleft <= 0 then
			begin
				Timer.Enabled := false;
				Close;
			end
			else
			begin
        if pmode = 1 then
        begin
  				case Rand(5) of
  				0:	codestr := 'TU GL CQ TEST DE '+call;
  				1:	codestr := 'RR TU '+callstr;
  				2:	codestr := 'TU '+call;
  				3:	codestr := 'RR GL CQ TEST '+call;
  				4:	codestr := 'RR TU '+call+' TEST';
  				end;
  				codeptr := 1;
  				mode := calling;
          txReply.Caption := 'Serial >';
        end
        else
        begin
  				case Rand(4) of
  				0:	codestr := 'TU GL        ';
  				1:	codestr := 'RR TU        ';
          2:  codestr := 'TU        ';
  				3:	codestr := 'RR GL'+call+'        ';
  				end;
  				codeptr := 1;
  				mode := closing;
          txReply.Caption := '';
        end;
				StartSending;
			end;
		end
		else if Pos('?',edReply.Text) > 0 then
		begin
			s := edReply.Text;
			Delete(s,Pos('?',edReply.Text),1);
			edReply.Text := s;
      edReply.SelStart := Length(s);
      edReply.SelLength := 0;
			StopSending;
			codestr := exch2+' ';
			mode := inqso;
			codeptr := 1;
      txReply.Caption := 'Serial >';
			StartSending;
		end
		else if Length(edReply.Text) = Length(exchange) then
		begin
			codestr := '  NR IS '+exchange+' '+exchange+'  QSL? ';
			mode := inqso;
			codeptr := 1;
      txReply.Caption := 'Serial >';
			StartSending;
		end
		else if counter = 0 then
		begin
			if Rand(3) > 0 then
			begin
				codestr := exch2+' '+exch2+' ';
				mode := inqso;
				codeptr := 1;
        txReply.Caption := 'Serial >';
				StartSending;
			end
			else
			begin
				codestr := callstr;
				codeptr := 1;
				mode := calling;
        txReply.Caption := 'Call >';
				StartSending;
			end;
		end;
	end;
  if pmode = 2 then
  begin
    case mode of
    calling:  edReply.Text := call;
    inqso:    edReply.Text := exchange;
    end;
  end;
end;

procedure TMorseMain.TuneClick(Sender: TObject);
var
	i,s: Integer;

  function RandChar( min, max: char): char;
  var
  	m,r: Integer;
  begin
  	m := Ord(min);
  	r := Ord(max) - m + 1;
  	Result := Chr(Rand(r)+m);
  end;

begin
	StopSending;
  if cbMode.Enabled then
  begin
    pmode := cbMode.ItemIndex;
    cbMode.Enabled := false;
  end;
	f := sePitch.Value - 49 + Rand(100);
	s := seSpeed.Value - 1 + Rand(3);
	do_sine_table;
	dotlength := ((sample_rate * 12) div 10) div s;	edReply.SetFocus;
	repeat
		call := calls[Rand(numcalls)];
		if Pos('/',call) < (Length(call)-3) then
		begin
			if call[Length(call)] in ['A'..'Z'] then
				call[Length(call)] := RandChar('A','Z');
			if call[Length(call)-1] in ['A'..'Z'] then
				call[Length(call)-1] := RandChar('A','Z');
			if call[Length(call)-2] in ['A'..'Z'] then
				call[Length(call)-2] := RandChar('A','Z');
		end;
	until CompareText(call,edUrCall.Text) <> 0;
  if pmode = 1 then
  begin
  	case Rand(6) of
  	0:	callstr := Format('TEST %s %s',[call,call]);
  	1:  callstr := Format('CQ TEST %s %s',[call,call]);
  	2:  callstr := Format('%s %s TEST',[call,call]);
  	3:  callstr := Format('TEST %s',[call]);
  	4:  callstr := Format('TEST %s %s TEST',[call,call]);
  	5:  callstr := Format('CQ CQ TEST %s %s',[call,call]);
  	end
  end
  else
  begin
    case Rand(6) of
    0, 1, 2,
    3:  callstr := call;
    4, 5:
        callstr := Format('%s %s',[call,call]);
    6:  callstr := Format('%s de %s %s K',[edUrCall.Text,call,call]);
    end;
  end;
	exchange := Copy(IntToStr(1001+Rand(Trunc((Now-timestarted)*8640))),2,6);
	exchangestr := edUrCall.Text;
	case Rand(3) of
	0:	exch2 := ' 599 '+exchange;
	1:	exch2 := ' 5NN '+exchange;
	2:	begin
				exch2 := ' 5NN '+exchange;
				i := 6;
				while i < 9 do
				begin
					if exch2[i] = '0' then exch2[i] := 'T';
					inc(i);
				end;
				i := 6;
				while i < 9 do
				begin
					if exch2[i] = '9' then exch2[i] := 'N';
					inc(i);
				end;
			end;
	end;
	case Rand(7) of
	0:	exchangestr := 'RR'+exch2+exch2;
	1:	exchangestr := 'TU'+exch2+exch2;
	2:	exchangestr := exchangestr+' RR'+exch2+exch2;
	3:	exchangestr := exchangestr+' TU'+exch2+exch2;
	4:	exchangestr := exchangestr+' RR'+exch2+exch2+' BK';
	5:	exchangestr := exchangestr+exch2+exch2;
	6:	exchangestr := exchangestr+exch2+exch2+' #';
	end;
	mode := listeningforcall;
  txReply.Caption := 'Call >';
	counter := 0;
end;

procedure TMorseMain.VolumeChange(Sender: TObject);
begin
	do_sine_table;
end;

procedure TMorseMain.WaveOutProc(hwo: HWAVEOUT; uMsg: UINT; dwParam1, dwParam2: DWORD);
var
  free_header: pWaveHdr;
begin
  case uMsg of
    WOM_OPEN  : ; // this message is never received
    WOM_CLOSE : ; // this message is never received
    WOM_DONE  : begin
                  // handle the wave out done message by writing the next buffer, if required
                	Inc(chars_played);
                  // point to wave header just completed, i.e. the next free buffer
                  free_header := pWaveHdr(dwParam1);
                  waveOutUnprepareHeader(hWave_out, free_header, SizeOf(WaveHdr));
                  if codeptr > Length(codestr) then
                  begin
                    // everything written has been played
                    // release the wave headers
                    if mode = calling then
                    begin
                      mode := listeningforcall;
                      counter := 4;
                    end
                    else if mode = inqso then
                    begin
                      mode := listeningforexch;
                      counter := 10;
                    end
                    else if mode = closing then
                    begin
                      mode := tuning;
                      counter := 1;
                    end;
                    key_down := false;
              		end
                  else
                  begin
                    // and write the next buffer, re-using the one just played
                    write_next_buffer(free_header);
              		end
                end;
  end
end;

procedure TMorseMain.write_next_buffer (header: pWaveHdr);
begin
  with header^ do
		dwBufferLength := fill_buffer_with_char (pBuffer(lpData), codestr[codeptr]);
	inc(codeptr);
  // write the buffer and bump the number written
  waveOutPrepareHeader (hWave_out, header, SizeOf(WaveHdr));
  waveOutWrite (hWave_out, header, SizeOf(WaveHdr));
	Inc(chars_sent);
end;

procedure TMorseMain.do_sine_table;
var
	i: 0..sine_table_samples - 1;
	y: extended;
begin
	vol := tbVolume.Position / (8.0 + Rand(8));
	for i := 0 to sine_table_samples - 1 do
	begin
		y := round (vol * 32767.0 * sin (2.0* i * Pi / sine_table_samples));
		sine_table^ [i] := round (y);
	end;
end;

procedure TMorseMain.fill_buffer_with_sinewave (bfr: pBuffer;  var index, samples: integer);
const
  fract_bits = 15;
var
  sample: audio_sample;
  d_angle: integer;      // 32-bit number, with 14 fractional bits, i.e. 17.15
	i,j,max_angle: integer;
begin
	angle := 0;
  // compute the angular step per sample corresponding to the desired frequency
	d_angle := round ((sine_table_samples shl fract_bits) * f / sample_rate);
  // this is the maximum number of samples in the sine table
	max_angle := (sine_table_samples shl fract_bits) - 1;
  for i := 1 to samples do
  begin
  	sample := sine_table^[angle shr fract_bits];// get current sine value
    if i < 32 then
			for j := i to 32 do
      	sample := (sample * 8) div 9;
    if i > (samples - 32) then
			for j := samples-32 to i do
				sample := (sample * 8) div 9;
		bfr^ [index] := sample;
    Inc (index);                               // bump the buffer pointer
    Inc (angle, d_angle);                      // bump the angle
		angle := angle and max_angle;              // wrap to 360 degrees
  end;
end;

procedure TMorseMain.fill_buffer_with_silence (bfr: pBuffer;  var index, samples: integer);
var
  sample: integer;
begin
  for sample := 0 to samples - 1 do
	begin
    bfr^ [index] := 0;                         // store it in the caller's buffer
    Inc (index);                               // bump the buffer pointer
  end;
end;

function TMorseMain.fill_buffer_with_char (bfr: pBuffer; ch: char): integer;
// This procedure fills a  buffer with a Morse character.
var
  sample, chunk_samples, c, i: integer;
	mc: string;
begin
	sample := 0;
  c := Ord(ch);
  if (c >= 32) and (c <= 90) then
  begin
  	mc := code[c];
    i := 1;
    while i <= Length(mc) do
    begin
    	case mc[i] of
      '.':	begin
							chunk_samples := dotlength;
			      	fill_buffer_with_sinewave(bfr, sample, chunk_samples);
							fill_buffer_with_silence(bfr, sample, dotlength);
						end;
			'-':	begin
							chunk_samples := dotlength * 3;
							fill_buffer_with_sinewave(bfr, sample, chunk_samples);
							fill_buffer_with_silence(bfr, sample, dotlength);
						end;
			else
				chunk_samples := dotlength * 2;
				fill_buffer_with_silence(bfr, sample, chunk_samples);
			end;
			Inc(i);
		end;
	end;
	chunk_samples := dotlength * (2+spacing);
	fill_buffer_with_silence(bfr, sample, chunk_samples);
	Result := sample * 2; // no of bytes (2 per sample)
end;

initialization
  {$I mainform.lrs}

end.

